-------------------------------------------------------------
--
--	     
--
--	pathfind.lua
--
--	ver.2012/02/12
--
-------------------------------------------------------------
--   AI.lua
-- require "./AI/USER_AI/pathfind.lua"
--
-- TagCell (x,y)		--      ""
-- ClearCells ()		--  
-- CalcStep (x,y,destx,desty)	--    
-- FoundFarawayPos (x,y)	--     x,y 
--
--      x,y  destx,desty, 
--  CalcStep()  nil
--

MAP_ENABLED = 1
MAP_PROBABLE = 2		
MAP = {}		--   ( "" )


--  
function ClearCells ()
	MAP = {}
end


--     nx,ny  ""
function TagCell (nx,ny,tag)
	if not MAP[ny] then
		MAP[ny] = {}
	end

	if tag ~= nil then
		MAP[ny][nx] = tag
	else
		MAP[ny][nx] = MAP_ENABLED
	end
end


--        nx,ny ""  
--        
function isValidCell (nx,ny,tag,maxdist)
--	TraceAI (string.format ("In isValidCell md=%d nx,ny=%d,%d",md,nx,ny))

--	local OwnerX,OwnerY = GetV (V_POSITION,MyOwner)

	local md = 14
	if (maxdist ~= nil) then
		md = maxdist
	end
	
	if (md < GetDistance(nx,ny,OwnerX,OwnerY)) or
	not (MAP[ny] and MAP[ny][nx] and MAP[ny][nx] <= tag) then
		return false
	end

	return true
end


--      x1,y1 - x2,y2 ""
function isValidLine (x1,y1,x2,y2,tag)
--	TraceAI (string.format ("In isValidLine x1,y1=%d,%d x2,y2=%d,%d",x1,y1,x2,y2))

	local len = GetDistance (x1,y1,x2,y2)
	local dx,dy = (x1-x2)/len,(y1-y2)/len

	local i = 0
	while i <= len do
		if not isValidCell (math.floor(x1-i*dx),math.floor(y1-i*dy),tag) then
			return false
		end

		i = i + 1
	end

	return true
end


--  "" ,   ,  mark
function MarkValidNeibours (map,nx,ny,mark,tag,maxdist)
	local result = false

	function MarkCell (map,mx,my,mark)
		if not map[my] then
			map[my] = {}
		end

		map[my][mx] = mark
		return true
	end

	function isMarkedCell (map,mx,my)
		if not (map and map[my] and map[my][mx]) then
			return false
		end
		return true
	end

	local neibours = {
		{ nx, ny + 1 },
		{ nx + 1, ny + 1 },
		{ nx + 1, ny },
		{ nx + 1, ny - 1 },
		{ nx, ny - 1 },
		{ nx - 1, ny - 1 },
		{ nx - 1, ny },
		{ nx - 1, ny + 1 },
	}

	local i,v
	for i,v in ipairs (neibours) do
--		TraceAI (string.format ("In MarkValidNeibours i=%d v=%d,%d",i,v[1],v[2]))

		if isValidCell (v[1],v[2],tag,maxdist) and not isMarkedCell (map,v[1],v[2]) then
			result = MarkCell (map,v[1],v[2],mark) or result
--			TraceAI (string.format ("In MarkValidNeibours result=%s map[%d][%d]=%s",tostring(result),v[1],v[2],tostring(map[v[2]] and map[v[2]][v[1]])))
		end
	end

	return result
end



--     
function FoundFarawayPos (x,y)
	local tempmap = {}
	tempmap[y] = {}
	tempmap[y][x] = 0

	local i,isvalid = 0,true
	local nx,ny,row,v

	--     
	while (isvalid) do
		isvalid = false

		local tagcount,tagpos = 0,{}

		for ny,row in pairs (tempmap) do
			for nx,v in pairs (row) do
				if (v == i) then
					tagpos[tagcount] = {nx, ny}
					tagcount = tagcount + 1
				end
			end
		end

		for tagcount,v in pairs (tagpos) do
--			TraceAI (string.format ("In FoundFarawayPos nx,ny=%d,%d i=%d",v[1],v[2],i))

			--     ,     ""
			isvalid = MarkValidNeibours (tempmap,v[1],v[2],i+1,MAP_ENABLED,100500) or isvalid
		end

		if (not isvalid) then
			nx,ny = x,y
			local dist = 0

			for tagcount,v in pairs (tagpos) do
				local d = GetDistance (x,y,v[1],v[2])

				if (d > dist) then
					d = dist
				        nx,ny = v[1],v[2]
				end
			end

			return nx,ny
		end

		i = i + 1
	end
end


--         x,y   destx,desty
function CalcStep (x,y,destx,desty,d,maxdist,tag)
	if (GetDistance (x,y,destx,desty) <= 1) then
		return destx,desty
	end

	local tempmap = {}
	tempmap[desty] = {}
	tempmap[desty][destx] = 0

	local i,isvalid,isfound = 0,true,false
	local nx,ny,row,v
	local vtag = MAP_ENABLED

	--  tag ,     "" 
	if tag ~= nil then
		vtag = tag             
	end

--	TraceAI ("In CalcStep 1")
	--     
	while (isvalid and not isfound) do
		isvalid = false

		local tagcount,tagpos = 0,{}

		for ny,row in pairs (tempmap) do
			for nx,v in pairs (row) do
				if (v == i) then
					tagpos[tagcount] = {nx, ny}
					tagcount = tagcount + 1
				end
			end
		end

		for tagcount,v in pairs (tagpos) do
--			TraceAI (string.format ("In CalcStep nx,ny=%d,%d i=%d",v[1],v[2],i))
			if (GetDistance (v[1],v[2],x,y) <= 1) then
				isfound = true
				break
			end

			--     ,     ""
			isvalid = MarkValidNeibours (tempmap,v[1],v[2],i+1,vtag,maxdist) or isvalid
		end

		if (not isvalid and not isfound and i < d) then	
			--     ,   
			--   d   
			for ny = desty+i,desty-i,-1 do
				for nx = destx-i,destx+i,1 do
					MarkValidNeibours (tempmap,nx,ny,i+1,vtag,maxdist)
				end
			end

			isvalid = true
		end

		i = i + 1
	end

	if (not isfound) then
		return
	end

--	TraceAI ("In CalcStep 2")
	--  ,    , ,  
	--      ,    "" 
	local stepx,stepy
	for ny,row in pairs (tempmap) do
		for nx,v in pairs (row) do
			if (tempmap[ny][nx] < i
			and GetDistance (x,y,nx,ny) <= MaxMoveDistance
			and isValidLine (x,y,nx,ny,vtag)) then
--				TraceAI (string.format ("In CalcStep: Best longest line is: x,y=%d,%d nx,ny=%d,%d l=%d",x,y,nx,ny,GetDistance(x,y,nx,ny)))
				i = tempmap[ny][nx]
				stepx,stepy = nx,ny
			end
		end
	end

	return stepx,stepy
end



function CalcRandStep (x,y,destx,desty)
	local nx,ny,l

	local len = GetDistance (x,y,destx,desty)
	local d = math.min (MaxMoveDistance, len)

	while (nx == nil or (GetDistance (x,y,nx,ny) > MaxMoveDistance)
		or 14 < GetDistance(nx,ny,OwnerX,OwnerY)) do

		--     x,y - destx,desty
		l = math.random (1, d)
		nx = x + math.floor (l * (destx-x) / len)
		ny = y + math.floor (l * (desty-y) / len)

		--    
		nx = nx + math.random (0, 2) - 1;
		ny = ny + math.random (0, 2) - 1;
	end

	return nx,ny
end


function LogMap (map,x,y,destx,desty)
	local l,r,u,d = OwnerX-14,OwnerX+14,OwnerY+14,OwnerY-14

	if l < 1 then l = 1; end
	if d < 1 then d = 1; end
	
	TraceAI (string.format("- %d %d ----------------------",l,u))
	local i,k,s
	for k = u,d,-1 do
		s = "|"
		for i = l,r,1 do
			if i == x and k == y then
				s = s .. "@"
			elseif i == destx and k == desty then
				s = s .. "X"
			elseif i == OwnerX and k == OwnerY then
				s = s .. "O"
			elseif map[k] and map[k][i] then
				s = s .. (math.mod (map[k][i], 10))
			else
				s = s .. " "
			end		
		end
		TraceAI (string.format("%s|",s))
	end
	TraceAI (string.format("---------------------- %d %d -",r,d))
end

